#ifdef HAVE_CONFIG_H
# include "config.h"     
#endif

// C/C++ includes
#include<math.h>
#include<iostream>
#include<fstream>
#include<vector>
#include<map>
#include<string>
//#include<ctime>

// DUNE includes
#include"dune/common/mpihelper.hh" // An initializer of MPI
#include"dune/common/exceptions.hh" // We use exceptions
#include<dune/common/fvector.hh>
#include<dune/common/static_assert.hh>
#include<dune/common/timer.hh>
#include<dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#include<dune/grid/io/file/gmshreader.hh>
//#include<dune/grid/yaspgrid.hh>
#include<dune/istl/bvector.hh>
#include<dune/istl/operators.hh>
#include<dune/istl/solvers.hh>
#include<dune/istl/preconditioners.hh>
#include<dune/istl/io.hh>
//#include<dune/istl/superlu.hh>

#include<dune/pdelab/common/function.hh>
#include<dune/pdelab/common/vtkexport.hh>
//#include<dune/pdelab/finiteelementmap/p0fem.hh>
//#include<dune/pdelab/finiteelementmap/p12dfem.hh>
//#include<dune/pdelab/finiteelementmap/pk2dfem.hh>
#include<dune/pdelab/finiteelementmap/pk3dfem.hh>
#include<dune/pdelab/finiteelementmap/q12dfem.hh>
#include<dune/pdelab/finiteelementmap/q22dfem.hh>
#include<dune/pdelab/finiteelementmap/q1fem.hh>
#include<dune/pdelab/finiteelementmap/p1fem.hh>
//#include<dune/pdelab/finiteelementmap/rannacher_turek2dfem.hh>
#include<dune/pdelab/finiteelementmap/conformingconstraints.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include<dune/pdelab/gridfunctionspace/genericdatahandle.hh>
#include<dune/pdelab/gridfunctionspace/interpolate.hh>
//OLD #include<dune/pdelab/gridfunctionspace/constraints.hh>
#include<dune/pdelab/constraints/constraints.hh>
#include<dune/pdelab/constraints/constraintsparameters.hh>
//OLD #include<dune/pdelab/gridoperatorspace/gridoperatorspace.hh>
#include<dune/pdelab/gridoperator/gridoperator.hh>
//#include<dune/pdelab/gridoperatorspace/gridoperatorspaceutilities.hh>
#include<dune/pdelab/backend/istlvectorbackend.hh>
#include<dune/pdelab/backend/istlmatrixbackend.hh>
#include<dune/pdelab/backend/istlsolverbackend.hh> //parallel
#include<dune/pdelab/stationary/linearproblem.hh>
#include<dune/pdelab/instationary/onestep.hh>


//VS includes
#include<dune/common/fmatrix.hh>
#include<dune/grid/uggrid.hh>
//Alugrid:
#include<dune/grid/alugrid.hh>
#include<dune/grid/io/file/dgfparser/dgfalu.hh>
#include<dune/grid/io/file/dgfparser/dgfparser.hh>

#include<dune/grid/io/file/dgfparser/dgfug.hh>
#include<dune/grid/utility/structuredgridfactory.hh>  // for uggrid creation
#include<dune/grid/uggrid/uggridfactory.hh>
#include<dune/pdelab/finiteelementmap/rt0q3dfem.hh>


//Parallel
#include<dune/pdelab/backend/novlpistlsolverbackend.hh>
#include<dune/grid/common/gridinfo.hh>

template<typename GV, typename RF>
class GaussianFunction : public Dune::PDELab::GridFunctionBase<Dune::PDELab::
           GridFunctionTraits<GV,RF,1,Dune::FieldVector<RF,1> >, GaussianFunction<GV,RF> >
{
const GV& gv;
public:
  typedef Dune::PDELab::GridFunctionTraits<GV,RF,1,Dune::FieldVector<RF,1> > Traits;

  GaussianFunction (const GV& gv_) : gv(gv_) {}
  // later: pass vector atoms and build initial density around !

  //! evaluate extended function on element
  inline void evaluate (const typename Traits::ElementType& e, 
                        const typename Traits::DomainType& xlocal,
                        typename Traits::RangeType& y) const
  {  
  	const int dim = Traits::GridViewType::Grid::dimension;
    typedef typename Traits::GridViewType::Grid::ctype ctype;
    Dune::FieldVector<ctype,dim> x = e.geometry().global(xlocal);
    
		if (x.two_norm() < 5)
    	y = exp(-(x.two_norm2()));  
		else   
    	y = 0.0;
  	return;
  }
  inline const GV& getGridView () {return gv;}
};    
    

int main(int argc, char** argv)
{
  try{
    //Maybe initialize Mpi
    Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
    std::cout << "Hello World! This is dune-interpolate." << std::endl;
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout<<"I am rank "<<helper.rank()<<" of "<<helper.size()
        <<" processes!"<<std::endl;
    
 
  MPI_Comm comm = helper.getCommunicator();
  
  if (helper.rank() == 0)
      {
        std::cout << "ALU-Grid from Tetgen" << std::endl;
      }  
      typedef Dune::ALUSimplexGrid<3,3> Grid;
      Dune::GridPtr<Grid> gridptr("tetgen_grid10.dgf",comm);
      
      Grid &grid = *gridptr;
      grid.globalRefine(2);
      //grid.loadBalance();
      
      typedef Dune::ALUSimplexGrid<3,3>::LevelGridView GV; 
      const GV& gv1=grid.levelView(1);
      //grid.globalRefine(1);
      const GV& gv2=grid.levelView(2);


      if (helper.rank() == 0)
      {
        Dune::gridinfo(grid);
        std::cout << "Load is Balanced" << std::endl;
      }  
 			Dune::VTKWriter<GV> vtkwriter(gv1,Dune::VTKOptions::conforming);
			Dune::SubsamplingVTKWriter<GV> vtksubwriter_level1(gv1,1);
			Dune::SubsamplingVTKWriter<GV> vtksubwriter_level2(gv2,1);


      typedef GV::Grid::ctype Coord;
			typedef double Real;
			typedef Dune::PDELab::Pk3DLocalFiniteElementMap< GV, Coord, Real, 2 > GV1P2FEM;
			GV1P2FEM GV1P2fem(gv1);
      typedef Dune::PDELab::Pk3DLocalFiniteElementMap< GV, Coord, Real, 2 > GV2P2FEM;
			GV2P2FEM GV2P2fem(gv2);
      
      typedef Dune::PDELab::NonoverlappingConformingDirichletConstraints CON;
			CON GV1P2con;
			CON GV2P2con;
			typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
			typedef Dune::PDELab::GridFunctionSpace<GV,GV1P2FEM,CON,VBE> GV1P2GFS;
			typedef Dune::PDELab::GridFunctionSpace<GV,GV2P2FEM,CON,VBE> GV2P2GFS;
			GV1P2GFS GV1P2gfs(gv1,GV1P2fem,GV1P2con);			
  		GV2P2GFS GV2P2gfs(gv2,GV2P2fem,GV2P2con);			
			//GV1P2con.compute_ghosts(GV1P2gfs);	
			//GV1P2con.compute_ghosts(GV2P2gfs);	
			typedef Dune::PDELab::BackendVectorSelector<GV1P2GFS,Real>::Type GV1P2GFVec;
			GV1P2GFVec GV1P2vec(GV1P2gfs,0.0);
      typedef Dune::PDELab::BackendVectorSelector<GV2P2GFS,Real>::Type GV2P2GFVec;
 			GV2P2GFVec GV2P2vec(GV2P2gfs,0.0);
			
      /*
      typedef P1GFS::ConstraintsContainer<Real>::Type P1CC;
      typedef P2GFS::ConstraintsContainer<Real>::Type P2CC;
			P1CC P1cc;
			P2CC P2cc;
			P1cc.clear();
			P2cc.clear();
			Dune::PDELab::constraints(b,P1gfs,P1cc,false); // Fill constraints container
			Dune::PDELab::constraints(b,P2gfs,P2cc,false); // Fill constraints container
			*/
      //1) Map from P1 to P2 function-space on same grid
      // 		- setup function e.g gaussian and interpolate to P1-vector
      //  	- create dgf-function from P1-vector
      //		- fct2gridfunction adapter
      //		- interpolate function to P2-vector
      //		- compare result to direct interpolation with P2-vector
      
			//Setup a function and calculate the coefficients vector from it
      typedef GaussianFunction<GV,Real> GaussF;
			GaussF gaussfunction(gv2);                 
			Dune::PDELab::interpolate(gaussfunction,GV2P2gfs,GV2P2vec);  

			// Take gfs and coefficients vector and transform them into a
			// grid function
			typedef Dune::PDELab::DiscreteGridFunction<GV1P2GFS,GV1P2GFVec> GV1P2DGF;
			typedef Dune::PDELab::DiscreteGridFunction<GV2P2GFS,GV2P2GFVec> GV2P2DGF;
			GV2P2DGF GV2P2dgf(GV2P2gfs,GV2P2vec);
			
		  std::cout << "GV2P2dgf defined" << std::endl;

			vtksubwriter_level2.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<GV2P2DGF>(GV2P2dgf,"Interpolated GV2P2 function"));
	    vtksubwriter_level2.write("GV2P2function_on_fine_grid",Dune::VTKOptions::ascii);
			vtksubwriter_level2.clear();

			std::cout << "vtk_level2 finished" << std::endl;
			
			
			//Dune::PDELab::GridFunctionToFunctionAdapter<GV2P2DGF> P2func(GV2P2dgf);
			//Dune::PDELab::interpolate(P2func,GV1P2gfs,GV1P2vec);   
			
			
			// man steckt gridfunction hinein und erhaelt dadurch eine function
			// d.h. es wird intern zunächst das zugehörige element zu x ermittelt um 
			// dann grid function auszuwerten
			// (gridfunction evaluiert ueber element, funktion hat direct functionswert)
			// im inversen Fall: FunctionToGridFunctionAdapter wird lediglich wrapper 
			// um die Funktion gelegt und diese wird nun
	    // ueber das element und lokale Koordinaten angesprochen
	    // d.h. definierend für gridfunction:
	    //inline void evaluate (const typename Traits::ElementType& e,
			//  const typename Traits::DomainType& x,typename Traits::RangeType& y) const

			std::cout << "Function defined" << std::endl;
						
			Dune::PDELab::interpolate(GV2P2dgf,GV1P2gfs,GV1P2vec);  
			std::cout << "Interpolate called" << std::endl;	
			
			GV1P2DGF GV1P2dgf(GV1P2gfs,GV1P2vec);
			
			std::cout << "GV1P2dgf defined" << std::endl;
			vtksubwriter_level1.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<GV1P2DGF>(GV1P2dgf,"Transformed GV2P2 function on GV1"));
      //vtkwriter.pwrite("P2function_output","./pics","./Test",Dune::VTKOptions::binary);
      vtksubwriter_level1.write("GV1P2function_on_rough_grid",Dune::VTKOptions::ascii);
 			
  return 0;    
  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
  } 
  
  
  
  
}
